# https://github.com/MagicStack/asyncpg
import asyncio
from typing import AsyncGenerator

__func_alias__ = {"channel_": "channel"}


def __virtual__(hub):
    return hub.acct.beacon.pg.HAS_LIBS


async def listen(hub, ctx) -> AsyncGenerator:
    if not ctx.connected:
        hub.log.warning("No server to listen on for pika plugin")
        return

    # Send the message to end all channels for this context
    def _done(conn):
        hub.log.info("A PostgreSQL beacon has disconnected")
        ctx.connected = False

    ctx.connection.add_termination_listener(_done)
    ctx.connection.add_log_listener(hub.beacon.pg.log)

    listeners = [hub.beacon.pg.channel(ctx, channel) for channel in ctx.channels]
    generator = hub.pop.loop.as_yielded(listeners)

    async for message in generator:
        yield message

    hub.log.debug("No more messages in PostgreSQL queue")


async def channel_(hub, ctx, channel: str) -> AsyncGenerator:
    """
    Add a listener for the PostgreSQL channel.
    Route the callback to a local queue and turn this function into a generator fo the results
    """
    queue = asyncio.Queue()

    async def _callback(conn, pid, ch, payload):
        await queue.put(payload)

    await ctx.connection.add_listener(channel, _callback)

    while hub.beacon.RUN_FOREVER and ctx.connected:
        yield await queue.get()

    hub.log.debug(f"No more messages in PostgreSQL channel {channel}")


def log(hub, conn, message):
    """
    A log listener for pg queues
    """
    hub.log.debug(message)



# https://aio-pika.readthedocs.io/en/latest/rabbitmq-tutorial/1-introduction.html
from typing import AsyncGenerator


def __virtual__(hub):
    return hub.acct.beacon.pika.HAS_LIBS


async def listen(hub, ctx) -> AsyncGenerator:
    if not ctx.connected:
        hub.log.warning("No server to listen on for pika plugin")
        return

    listeners = [hub.beacon.pika.channel(ctx, ch) for ch in ctx.channels]
    generator = hub.pop.loop.as_yielded(listeners)

    async for message in generator:
        async with message.process():
            yield message.body.decode()

    hub.log.debug("No more messages in ampq queue")


async def channel(hub, ctx, name: str) -> AsyncGenerator:
    c = await ctx.connection.channel()
    queue = await c.declare_queue(name)
    hub.log.debug(f"Connected to amqp server channel {name}")

    async with queue.iterator() as q_iter:
        hub.log.debug(f"Started ampq iterator on channel {name}")
        async for message in q_iter:
            yield message

    hub.log.debug("No more messages in ampq queue")



from dict_tools import data
from typing import Any, Dict


async def post_gather(hub, ctx) -> Dict[str, Any]:
    return data.NamespaceDict(ctx.ret)


async def post_ctx(hub, ctx):
    return data.NamespaceDict(ctx.ret)



import asyncio
from typing import AsyncGenerator, Any, Dict, List

STOP_ITERATION = object()


def __init__(hub):
    hub.beacon.STARTED = False
    hub.beacon.RUN_FOREVER = True
    hub.pop.sub.add(dyne_name="evbus")
    hub.pop.sub.load_subdirs(hub.beacon)


def cli(hub):
    """
    Start up on the cli and print out the returns as they come
    """
    hub.pop.config.load(["beacon", "evbus", "acct", "rend"], cli="beacon")
    hub.pop.loop.create()

    beacon_profiles = hub.pop.Loop.run_until_complete(
        hub.evbus.init.profiles(
            subs=["beacon"],
            acct_file=hub.OPT.acct.acct_file,
            acct_key=hub.OPT.acct.acct_key,
            acct_profiles=hub.OPT.evbus.ingress_profiles,
        )
    )
    evbus_profiles = hub.pop.Loop.run_until_complete(
        hub.evbus.init.profiles(
            subs=["evbus"],
            acct_file=hub.OPT.acct.acct_file,
            acct_key=hub.OPT.acct.acct_key,
            acct_profiles=hub.OPT.evbus.ingress_profiles,
        )
    )
    coros = [
        hub.pop.Loop.create_task(
            hub.beacon.init.start(
                format_plugin=hub.OPT.beacon.format, sub_profiles=beacon_profiles
            )
        ),
        hub.pop.Loop.create_task(hub.evbus.init.start(evbus_profiles)),
        hub.evbus.init.print(renderer=hub.OPT.rend.output),
    ]
    if hub.OPT.evbus.random:
        coros.append(hub.evbus.test.start())

    try:
        for task in asyncio.as_completed(coros, loop=hub.pop.Loop):
            hub.pop.Loop.run_until_complete(task)
    except KeyboardInterrupt:
        hub.log.debug("Shutting down beacons")


async def start(hub, format_plugin: str, sub_profiles: Dict[str, Dict[str, Any]]):
    """
    Listen to all the beacon plugins, process events as they come in and vomit processed events to egress plugins
    """
    listeners_ = hub.beacon.init.listeners(sub_profiles)
    generator = hub.pop.loop.as_yielded(listeners_)

    # Ready to start a loop of generators
    hub.beacon.STARTED = True
    await asyncio.sleep(0)

    async for event, ref in generator:
        if event is STOP_ITERATION:
            hub.log.debug("Iterator was forced to stop")
            return

        processed = hub.beacon.format[format_plugin].apply(event, ref)
        await hub.evbus.BUS.put(processed)


async def join(hub):
    """
    Block until the event bus and beacons are started
    """
    await hub.evbus.init.join()

    while not hub.beacon.STARTED:
        await asyncio.sleep(0)


async def stop(hub):
    hub.log.debug("Shutting down beacons")
    # Stop beacons
    hub.beacon.RUN_FOREVER = False
    hub.evbus.RUN_FOREVER = False
    await hub.beacon.local.stop()
    await hub.evbus.init.stop()


def listeners(hub, contexts: Dict[str, Any]) -> List[AsyncGenerator]:
    ret = []

    hub.log.debug("Starting beacon listeners")
    for plugin in hub.beacon:
        name = plugin.__name__
        # If the listen function of a beacon has ctx, then populate it from acct
        if "ctx" in plugin.listen.signature.parameters:
            # Iterate over the defined profiles for the plugin
            for profile, ctx in contexts.get(name, {}).items():
                # Inject the ctx from the profile into the listener
                hub.log.debug(f"Created {name} beacon with ctx")
                ret.append(plugin.listen(ctx))
        else:
            hub.log.debug(f"Created {name} beacon")
            ret.append(plugin.listen())

    return ret


async def listen(hub) -> AsyncGenerator:
    """
    Listen for the stop signal then break out and shutdown everything
    """
    while hub.beacon.RUN_FOREVER:
        await asyncio.sleep(1)
    yield STOP_ITERATION



from typing import AsyncGenerator


async def sig_listen(hub, ctx) -> AsyncGenerator:
    ...


async def call_listen(hub, ctx):
    # Inject the ctx.ref into the yield results
    async for ret in ctx.func(*ctx.args, **ctx.kwargs):
        yield ret, ctx.ref



CLI_CONFIG = {
    "beacon_profiles": {"nargs": "*"},
    "format": {},
    # acct options
    "acct_file": {"source": "acct", "os": "ACCT_FILE"},
    "acct_key": {"source": "acct", "os": "ACCT_KEY"},
    # evbus options
    "ingress_profiles": {"source": "evbus"},
    "random": {"source": "evbus", "action": "store_true"},
    # rend options
    "output": {"source": "rend"},
}
CONFIG = {
    "beacon_profiles": {
        "help": "The acct profile names to allowlist.  If none are specified then all profiles will be used",
        "default": [],
    },
    "format": {
        "type": str,
        "help": "The plugin to use for processing data",
        "default": "json",
    },
}
DYNE = {
    "acct": ["acct"],
    "beacon": ["beacon"],
}



try:
    import aio_pika

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


async def gather(hub, profiles):
    """
    Get profile names from encrypted amqp credential files

    Example:
    .. code-block:: yaml

        pika:
          profile_name:
            host: localhost
            port: 5672
            beacon_channels:
              - channel1
              - channel2
    """
    sub_profiles = {}
    for profile, ctx, in profiles.get("pika", {}).items():
        sub_profiles[profile] = {
            "connected": False,
            "channels": ctx.pop("beacon_channels", []),
            "connection": None,
        }

        hub.log.debug("connecting to rabbitmq server")
        try:
            sub_profiles[profile]["connection"] = await aio_pika.connect_robust(
                loop=hub.pop.Loop, **ctx
            )
            sub_profiles[profile]["connected"] = True
            hub.log.debug("connected to rabbitmq server")
        except Exception as err:
            hub.log.error("Could not connect to rabbitmq server")
            hub.log.error(str(err))

    return sub_profiles



import datetime
import json
from typing import Any, Dict


def apply(hub, event: Dict, ref: str) -> Dict[str, Any]:
    # Make sure the event is a json serializable object
    try:
        event = json.dumps(event, sort_keys=True)
    finally:
        event = json.loads(event)

    return {"data": event, "timestamp": datetime.datetime.now().timestamp(), "ref": ref}



import asyncio
from typing import AsyncGenerator


def __init__(hub):
    hub.beacon.local.QUEUE = asyncio.Queue()


STOP_ITERATION = object()


async def listen(hub) -> AsyncGenerator:
    """
    listen for data on the local queue
    """
    async for event in hub.beacon.local.channel():
        yield event


async def channel(hub) -> AsyncGenerator:
    while hub.beacon.RUN_FOREVER:
        event = await hub.beacon.local.QUEUE.get()
        if event is STOP_ITERATION:
            return
        yield event

    hub.log.debug("No more messages in local queue")


async def stop(hub):
    await hub.beacon.local.QUEUE.put(STOP_ITERATION)
